from shard.routers.base import BaseRouter
from shard.mixins import SpecificDatabaseMixin


class SpecificRouter(BaseRouter):
    def db_for_read(self, model, **hints):
        return self._get_specific_database(model=model, **hints)

    def db_for_write(self, model, **hints):
        return self._get_specific_database(model=model, **hints)

    def _get_specific_database(self, model, **hints):
        if not issubclass(model, SpecificDatabaseMixin):
            return None

        return model.specific_database()



from typing import Optional

from shard.mixins import ShardMixin

from shard.routers.base import BaseReplicationRouter
from shard.utils.shard import get_shard_by_instance, get_shard_by_shard_key_and_shard_group


class ShardRouter(BaseReplicationRouter):
    def _get_master_database(self, model, **hints) -> Optional[str]:
        if not issubclass(model, ShardMixin):
            return None

        shard = None
        if hints.get('shard_key'):
            shard = get_shard_by_shard_key_and_shard_group(shard_key=hints['shard_key'], shard_group=model.shard_group)
        elif hints.get('instance'):
            shard = get_shard_by_instance(instance=hints['instance'])

        return shard



from typing import Optional

from shard.strategy.routing.random import RandomReadStrategy


class BaseRouter:
    def db_for_read(self, model, **hints):
        raise NotImplementedError

    def db_for_write(self, model, **hints):
        raise NotImplementedError


class BaseReplicationRouter(BaseRouter):
    _strategy = RandomReadStrategy

    def db_for_read(self, model, **hints):
        master = self._get_master_database(model=model, **hints)

        if master is None:
            return None

        return self._get_slave_database(master=master)

    def db_for_write(self, model, **hints):
        return self._get_master_database(model=model, **hints)

    def _get_master_database(self, model, **hints) -> Optional[str]:
        raise NotImplementedError

    def _get_slave_database(self, master: str) -> str:
        return self._strategy.pick_read_db(master=master)



# flake8: noqa: W0212 pylint: disable=protected-access
from django.db.models.query import QuerySet

__all__ = ('ShardQuerySet', )


class ShardQuerySet(QuerySet):
    def __init__(self, model, *args, **kwargs):
        super().__init__(model=model, *args, **kwargs)
        self._instance = None
        self._exact_lookups = {}

    def _clone(self):
        clone = super()._clone()
        clone._exact_lookups = self._exact_lookups.copy()
        return clone

    def _filter_or_exclude(self, negate, *args, **kwargs):
        clone = super()._filter_or_exclude(negate, *args, **kwargs)
        if getattr(clone, '_exact_lookups', None) is None:
            clone._exact_lookups = {}
        clone._exact_lookups.update(  # flake8: noqa: E1101 pylint: disable=no-member
            dict([(k, v) for k, v in kwargs.items() if '__' not in k])
        )
        return clone

    @property
    def db(self):
        if self._db:
            return self._db

        self._hints['exact_lookups'] = self._exact_lookups
        if not self._hints.get('instance') and getattr(self, '_instance', None):
            self._hints['instance'] = getattr(self, '_instance')

        return super().db

    def create(self, **kwargs):
        self._instance = self.model(**kwargs)
        return super().create(**kwargs)

    def get_or_create(self, defaults=None, **kwargs):
        defaults = defaults or {}
        lookup, _ = self._extract_model_params(defaults, **kwargs)
        self._exact_lookups = lookup
        return super(ShardQuerySet, self).get_or_create(defaults=defaults, **kwargs)

    def update_or_create(self, defaults=None, **kwargs):
        defaults = defaults or {}
        lookup, _ = self._extract_model_params(defaults, **kwargs)
        self._exact_lookups = lookup
        return super(ShardQuerySet, self).update_or_create(defaults=defaults, **kwargs)
